import csv
import gzip
import logging
import os
import subprocess
import tabix
import sys
import pandas as pd
import oncodrivemut.modules.lib as lib
import oncodrivemut.utils.constants as ct
from oncodrivemut.utils.progress import ProgressBar


def run(config: dict):
    """
    Retrive CADD scores

    :param config: The configuration dictionary
    :return: The configuration dictionary
    """

    logging.info('Retrieving CADD scores')

    # retrieve the paths of the outputs of the module
    output_fp = os.path.join(config['tmp_folder'], ct.CADD_OUTPUT)
    config['cadd_input_fp'] = os.path.join(config['tmp_folder'], ct.CADD_INPUT)
    config['cadd_output_fp'] = output_fp

    # assess if i need to run this stage by checking whether the *final* outputed file already exists
    if not config['force_output'] and os.path.isfile(output_fp):
        logging.debug('Stage skipped: file {0} already exists'.format(output_fp))
        return config

    # if the input format is a vcf, i can pass it directly to cadd after zipping it
    if config['input_format'] == ct.VCF_FORMAT:
        vcf_df = pd.read_csv(config['input_fp'], sep='\t')

        if not config['extended']:
            l = len(vcf_df)
            vcf_df = lib.filter_pams(vcf_df, config)
            logging.debug('{0} out of {1} entries are not PAM and not considered to retrieve CADD'.format(l-len(vcf_df), l))

        vcf_df.to_csv(config['cadd_input_fp'], index=False, sep='\t', compression='gzip')
        logging.debug('Input temp file for CADD copied and zipped from {0} to {1}'.format(config['input_fp'], config['cadd_input_fp']))

    # if not a vcf, i have to create a vcf like file to be inputed to cadd
    else:
        mutations_df = pd.read_csv(config['annotated_mutations_fp'], sep='\t')

        if not config['extended']:
            l = len(mutations_df)
            mutations_df = mutations_df[mutations_df[ct.CONSEQUENCE_TYPE].isin(ct.PAM_CT)]
            logging.debug('{0} out of {1} entries are not PAM and not considered to retrieve CADD'.format(l-len(mutations_df), l))

        # need to protect that no mutaions to check cadd are necessary available
        if mutations_df.empty:
            logging.debug('No mutation to be queried to CADD')
            return config

        # convert the mutations to a vcf needed for cadd
        vcf_df = lib.gdna_2_vcf(mutations_df, config)
        # this can be useful if the same mutation lead to two annotation (because it map to two genes, the one in +/- strands)
        vcf_df = vcf_df.drop_duplicates()

        # note that the cadd requires gzip, or it will run but with errors; note that i output the gdna entry to use it as
        # a key to recover the information
        vcf_df.to_csv(config['cadd_input_fp'], sep='\t', index=False, columns=[ct.CHR, ct.POS, 'ID', ct.REF, ct.ALT, ct.GDNA], compression='gzip')
        logging.debug('Input temp file for CADD created {0}'.format(config['cadd_input_fp']))

    fast_cadd(config)

    logging.debug('Output tmp file with CADD scores created {0}'.format(config['cadd_output_fp']))

    return config


def fast_cadd(config):

    cadd_scores_tabix = tabix.open(config['cadd_scores'])

    in_file, out_file = config['cadd_input_fp'],  config['cadd_output_fp']

    with gzip.open(in_file, 'rt') as in_fd, gzip.open(out_file, 'wt') as out_fd:

        writer = csv.writer(out_fd, delimiter='\t')
        writer.writerow(['## Fast CADD 20160418'])
        writer.writerow(['#CHROM', 'POS', 'REF', 'ALT', 'RawScore', 'PHRED'])

        reader = csv.reader(in_fd, delimiter='\t')
        header = next(reader)

        todo_rows = []

        for m in reader:
            chrom, pos, ref, alt = m[0], int(m[1]), m[3], m[4]

            if len(ref) == 1 and len(alt) == 1 and ref in "ACTG" and alt in "ACTG":
                try:
                    for r in cadd_scores_tabix.query(chrom, pos - 1, pos):
                        if r[2] == ref and r[3] == alt:
                            writer.writerow([r[0], r[1], r[2], r[3], r[4], r[5]])
                except tabix.TabixError:
                    logging.warning("Tabix error at {}".format(m))
                    todo_rows.append(m)
            else:
                todo_rows.append(m)

        if len(todo_rows) > 0:
            tmp_in_file = os.path.join(os.path.dirname(in_file), "cadd_input_tmp.tsv.gz")
            tmp_out_file = os.path.join(os.path.dirname(out_file), "cadd_output_tmp.tsv.gz")

            with gzip.open(tmp_in_file, 'wt') as tmp_fd:
                tmp_writer = csv.writer(tmp_fd, delimiter='\t')
                tmp_writer.writerow(header)
                for r in todo_rows:
                    tmp_writer.writerow(r)

            # run cadd
            cmd = " ".join([
                'bash',
                config['cadd_sh'],
                tmp_in_file,
                tmp_out_file
            ])

            progress_bar = ProgressBar(message_prefix="Running CADD ")
            progress_bar.start()

            with subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as proc:
                while True:
                    line = proc.stdout.readline().decode()
                    if line == '':
                        break
                    if progress_bar.debug:
                        progress_bar.set_message(line.strip())

                ret_code = proc.returncode

            progress_bar.stop()

            if ret_code:
                logging.error("Retcode running CADD {0} for command {1}".format(ret_code, cmd))
                sys.exit(-1)

            with gzip.open(tmp_out_file, 'rt') as tmp_fd:
                tmp_reader = csv.reader(tmp_fd, delimiter='\t')
                tmp_header = next(tmp_reader)
                tmp_header = next(tmp_reader)

                for r in tmp_reader:
                    writer.writerow([r[0], r[1], r[2], r[3], r[4], r[5]])

            try:
                os.unlink(tmp_in_file)
                os.unlink(tmp_out_file)
            except OSError as e:
                logging.debug("Error removing temp files. {}".format(str(e)))


def slow_cadd(config):
    # run cadd
    cmd = " ".join([
        'bash',
        config['cadd_sh'],
        config['cadd_input_fp'],
        config['cadd_output_fp']
    ])  # no redirection?
    logging.debug('Executing command: {0}'.format(cmd))
    progress_bar = ProgressBar(message_prefix="Running CADD ")
    progress_bar.start()
    with subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as proc:
        while True:
            line = proc.stdout.readline().decode()
            if line == '':
                break
            if progress_bar.debug:
                progress_bar.set_message(line.strip())

        ret_code = proc.returncode
    progress_bar.stop()
    if ret_code:
        logging.error("Retcode running CADD {0} for command {1}".format(ret_code, cmd))
        sys.exit(-1)

