__author__ = 'david'

import logging
import os
import sys

import numpy as np
import pandas as pd

import oncodrivemut.utils.constants as ct


def known_split_in_dna_and_prot(df, label):

    def collapse_values(tumor_se):

        if len(tumor_se) > 1:
            tumors = []
            for t_e in tumor_se:
                t_s = t_e.split(';') if ';' in t_e else [t_e]
                for t in t_s:
                    tumors.append(t)
            return ';'.join(list(set(tumors)))

        else:
            return tumor_se

    # select those entries in which the protein change is or is not stated (and collapse if the same protein change is stated more than once)
    dna_df = df[df[ct.PROT].isin(['.'])].copy()
    dna_df.loc[:, 'key'] = dna_df[ct.GDNA]
    prot_df = df[np.invert(df.index.isin(dna_df.index))]
    prot_df = prot_df[[ct.GENE, ct.PROT, label]].groupby([ct.GENE, ct.PROT]).aggregate(lambda values: collapse_values(values)).reset_index()
    prot_df['key'] = prot_df[ct.PROT]

    return dna_df.append(prot_df)


def match_known_oncogenic(mutation_df, run_info_d):

    def replace_is_known_inframe(row):

        cond = (row[ct.GENE] == 'EGFR') & ((row[ct.CONSEQUENCE_TYPE] in ct.INFRAME_CT and row[ct.EXON] == 19) |
                                           (row[ct.CONSEQUENCE_TYPE] == ct.INFRAME_CT[1] and row[ct.EXON] == 20))
        if cond:
            return 'L'

        return row[ct.KNOWN_ONCOGENIC]

    logging.debug('Matching protein/gDNA coordinates of the mutations with known oncogenic events..')

    known_oncogenic_df = pd.read_csv(run_info_d['known_oncogenic_fp'], sep='\t')

    for c in [ct.GDNA, ct.GENE, ct.PROT]:
        if c not in known_oncogenic_df.columns:
            logging.error('ERROR: The file {0} has no column named {1}!!'.format(run_info_d['known_oncogenic_fp'], c))
            sys.exit(-1)

    # rename columns and drop duplicates because in that file, a single mutation may map to two genes and thus has two entries
    # to_do: remove it in the metadata file, and not here
    known_oncogenic_df = known_oncogenic_df.rename(columns={'cancer_acronym': ct.KNOWN_ONCOGENIC})
    known_oncogenic_df = known_oncogenic_df.drop_duplicates([ct.GDNA, ct.KNOWN_ONCOGENIC])

    # divide the known mutations depending on they have protein changes (and then collapse to avoid duplicated protein changes) and dna
    known_oncogenic = known_split_in_dna_and_prot(known_oncogenic_df, ct.KNOWN_ONCOGENIC)

    # merge with the input file (note that i include the gene name because one mutation can map to two genes) by prot if availabe by dna if not
    mutation_df['key'] = np.where(mutation_df[ct.PROT].isin(['.']), mutation_df[ct.GDNA], mutation_df[ct.PROT])
    mutation_df = mutation_df.merge(known_oncogenic[[ct.GENE, ct.KNOWN_ONCOGENIC, 'key']], how='left', on=[ct.GENE, 'key'])

    # remove the tmp key
    mutation_df = mutation_df.drop(['key'], 1)

    # special condition for egfr inframe indels that are known but the current match script is not suited to detect them
    mutation_df[ct.KNOWN_ONCOGENIC] = mutation_df.apply(lambda row: replace_is_known_inframe(row), axis=1)

    return mutation_df


def match_known_predisposing(mutation_df, run_info_d):

    logging.debug('Matching protein/gDNA coordinates of the mutations with known predisposing variants..')

    known_predisposing_df = pd.read_csv(run_info_d['known_predisposing_fp'], sep='\t')

    for c in [ct.GDNA, ct.GENE, ct.PROT]:
        if c not in known_predisposing_df.columns:
            logging.error('ERROR: The file {0} has no column named {1}!!'.format(run_info_d['known_predisposing_fp'], c))
            sys.exit(-1)

    # rename columns and drop duplicates because in that file, a single mutation may map to two genes and thus has two entries
    # to_do: remove it in the metadata file, and not here
    known_predisposing_df = known_predisposing_df.rename(columns={'cancer_acronym': ct.KNOWN_PREDISPOSING})
    known_predisposing_df = known_predisposing_df.drop_duplicates([ct.GDNA, ct.KNOWN_PREDISPOSING])

    # divide the known mutations depending on they have protein changes (and then collapse to avoid duplicated protein changes) and dna
    known_predisposing_df = known_split_in_dna_and_prot(known_predisposing_df, ct.KNOWN_PREDISPOSING)

    # merge with the input file (note that i include the gene name because one mutation can map to two genes) by prot if availabe by dna if not
    mutation_df.loc[:, 'key'] = np.where(mutation_df[ct.PROT].isin(['.']), mutation_df[ct.GDNA], mutation_df[ct.PROT])
    mutation_df = mutation_df.merge(known_predisposing_df[[ct.GENE, ct.KNOWN_PREDISPOSING, 'key']], how='left', on=[ct.GENE, 'key'])

    # remove the tmp key
    mutation_df = mutation_df.drop(['key'],1)

    return mutation_df


def match_domains(mutation_df, config):

    def is_domain_overlapped(row):
        try:
            if int(row['Pfam start'] <= int(row[ct.AA_POS])) and int(row[ct.AA_POS]) <= int(row['Pfam end']):
                return 1
            return 0
        except:
            return 0

    logging.debug('Matching with Pfam domains by using protein positions..')

    if ct.TRANSCRIPT not in mutation_df.columns:
        logging.error('ERROR: The mutation file has no column named "transcript" as expected!!')
        sys.exit(-1)

    # read and format pfam file
    pfam_df = pd.read_csv(config['pfam_fp'], sep='\t')
    pfam_df = pfam_df.rename(columns={'Ensembl Transcript ID': ct.TRANSCRIPT, 'Pfam domain': ct.PFAM_DOMAIN})

    # integrate pfam data
    mutation_df = mutation_df.merge(pfam_df[[ct.TRANSCRIPT, 'Pfam start', 'Pfam end', ct.PFAM_DOMAIN]], how='left', on=ct.TRANSCRIPT)

    # for all the gene pfam domains, see if there is overlap with the protein change
    mutation_df['Pfam_overlap'] = mutation_df.apply(lambda row: is_domain_overlapped(row), axis=1)

    # this drop duplicates is necessary because i did the merge at the level of transcript, so all the mutation of a gene are duplicated by all the Pfam entries
    #  of that gene. Therefore, this dropduplciatessolves this issue by keeping that entry that overlaps (if any) with the pfam domain
    #  (note that the drop duplicates also take into account gdna, because vcf with several alt alleles
    #   creates several ouptuts with the same input, and transcript, becasue the same input can map to two genes)
    mutation_df = mutation_df.sort_values(by=['input', 'Pfam_overlap'], ascending=False) \
        .drop_duplicates(['input', ct.TRANSCRIPT, ct.GDNA, ct.INTERNAL_ID], keep='first')

    # recodify to have a filled domain field only if indeed there is match
    mutation_df[ct.PFAM_DOMAIN] = np.where(mutation_df['Pfam_overlap'] == 1, mutation_df[ct.PFAM_DOMAIN], '')

    # remove tmp columns
    mutation_df = mutation_df.drop(['Pfam start', 'Pfam end', 'Pfam_overlap'], 1)

    return mutation_df


def match_polymorphisms(mutation_df, run_info_d):

    logging.debug('Matching with polymorphisms rettrieved from Exac..')

    polym_df = pd.read_csv(run_info_d['polym_fp'], sep='\t')
    polym_df = polym_df.rename(columns={polym_df.columns[1]: ct.EXAC_AF})

    # merge
    mutation_df = mutation_df.merge(polym_df, on=ct.GDNA, how='left')

    return mutation_df


def match_transcript_metadata(mutation_df: pd.DataFrame, config: dict):

    logging.debug('Matching with transcript metadata..')

    transcript_df = pd.read_csv(config['transcript_metrics_fp'], sep='\t')
    transcript_df = transcript_df.rename(columns={'Ensembl Transcript ID': ct.TRANSCRIPT, 'transcript_exons': ct.TRANSCRIPT_EXONS,
                                                  'transcript_aminoacids': ct.TRANSCRIPT_AMINOACIDS})

    # merge
    mutation_df = mutation_df.merge(transcript_df[[ct.TRANSCRIPT, ct.TRANSCRIPT_AMINOACIDS, ct.TRANSCRIPT_EXONS]], how='left',
                                    on=ct.TRANSCRIPT)

    return mutation_df


def match_cadd(mutation_df: pd.DataFrame, config: dict):

    def create_cadd_key(row, cols):
        return '|'.join([str(row[cols[0]]), str(row[cols[1]]), row[cols[2]], row[cols[3]]])

    if not os.path.isfile(config['cadd_output_fp']):
        logging.debug('No CADD file available')
        mutation_df[ct.CADD_SCORE] = ''
        return mutation_df

    logging.debug('Matching with CADD scores obtained in a previous step..')
    cadd_output_df = pd.read_csv(config['cadd_output_fp'], sep='\t', skiprows=1)
    cadd_output_df = cadd_output_df.rename(columns={'PHRED': ct.CADD_SCORE})

    # if the input format is vcf, i can merge by the vcf coordinates
    if config['input_format'] == ct.VCF_FORMAT:
        cadd_output_df['input'] = cadd_output_df.apply(lambda row: create_cadd_key(row, cadd_output_df.columns), axis=1)
        mutation_df = mutation_df.merge(cadd_output_df[['input', ct.CADD_SCORE]], how='left', on='input')

    # if the input file is not a vcf, i need to check the input file passed to cadd to recover the original gdna key
    # to do the merge with the phred score in the cadd output file
    else:
        cadd_input_df = pd.read_csv(config['cadd_input_fp'], sep='\t')

        # merge the cadd files to have the gdna key (from the cadd input) and the phred score (from the cadd output) in a single file
        cadd_df = cadd_output_df.merge(cadd_input_df, how='outer', left_on=cadd_output_df.columns.tolist()[0:4],
                                       right_on=[cadd_input_df.columns.tolist()[i] for i in [0, 1, 3, 4]])

        # merge
        mutation_df = mutation_df.merge(cadd_df[[ct.GDNA, ct.CADD_SCORE]], how='left', on=ct.GDNA)

    cadd_na_df = mutation_df[mutation_df[ct.CADD_SCORE].isnull()]
    if len(cadd_na_df) > 0 and config['extended']:
        logging.warning('There are {0} entries with no CADD retrieved scores:'.format(len(cadd_na_df)))
        logging.debug('They are:\n{0}'.format(cadd_na_df[[ct.GDNA]]))

    return mutation_df


def run(config: dict):

    logging.info('Matching with several metadata files')

    # retrive the paths of the outputed files
    out_fp = os.path.join(os.path.join(config['tmp_folder'], ct.METADATA_MUTS))
    config['metadata_output_fp'] = out_fp

    # assess if i need to run this stage by checking whether the *final* outputed file already exists
    if not config['force_output'] and os.path.isfile(out_fp):
        logging.debug('Stage skipped: file {0} already exists'.format(out_fp))
        return config

    mutation_df = pd.read_csv(config['annotated_mutations_fp'], sep='\t')

    # match with transcript metrics metadata
    mutation_df = match_transcript_metadata(mutation_df, config)

    # match with known oncogenic mutations
    mutation_df = match_known_oncogenic(mutation_df, config)

    # match with known cancer predisposing variants
    mutation_df = match_known_predisposing(mutation_df, config)
    # match with polymorphisms
    mutation_df = match_polymorphisms(mutation_df, config)

    # match with pfam
    mutation_df = match_domains(mutation_df, config)

    # match with CADD (that have been obtained before)
    mutation_df = match_cadd(mutation_df, config)

    # output
    metadata_output_fp = config['metadata_output_fp']
    mutation_df.to_csv(metadata_output_fp, index=False, sep='\t')
    logging.debug('File with {0} entries created: {1}'.format(len(mutation_df), metadata_output_fp))

    return config
